"""Module with classes corresponding to VTK file types.

Currently, the following types are implemented:
- ImageData
- RectilinearGrid
- StructuredGrid
- UnstructuredGrid

The ParaViewData file format, although not part of VTK, is also available.

All instances of classes in this file are context managers and can be used in
the idiomatic:

  with RectilinearGrid(...) as file:
      file.addPointData(...)

"""

__copyright__ = "Copyright © 2018-2023 Lucas Frérot"
__license__ = "SPDX-License-Identifier: MIT"

import functools
import typing as ts
import numpy as np

from . import writer
from .data_array import DataArray
from .unstructured import CellType, check_connectivity


def _make_3darray(points):
    """Complete missing coordinates to get 3d points."""
    if points.shape[1] == 3:
        return points

    points_3d = np.zeros((points.shape[0], 3))
    for i in range(min(points.shape[1], 3)):
        points_3d[:, i] = points[:, i]
    return points_3d


def _fold_extent(extent, offsets, dimension):
    if offsets is None:
        offsets = [0] * len(extent)
    elif len(offsets) == dimension:
        offsets = list(offsets)
        offsets += [0] * (len(extent) - len(offsets))
    else:
        raise ValueError(
            'Size of offsets should '
            f'match domain dimension {dimension}')

    def fold_extent(acc, couple):
        offset, extent = couple
        offset -= offset != 0
        return acc + f"{offset} {offset+extent} "

    # Create extent string with offsets
    return functools.reduce(fold_extent, zip(offsets, extent), "")


class WriteManager:
    """Context manager that writes on exit."""

    def write(self):
        """Write file."""
        raise NotImplementedError()

    def __enter__(self):
        """Enter context."""
        return self

    def __exit__(self, err_type, *_):
        """Exit context and handle errors."""
        if err_type is None:
            # Only write if no error
            self.write()
        return True


class VTKFile(WriteManager):
    """Generic VTK file.

    Base class of all VTK-formated files. Sets up the common XML tree with
    PointData and CellData. Also supplies the functions to add point, cell and
    field data.

    """

    _FileDescriptor = writer.Writer.FileDescriptor

    def __init__(self, filename: _FileDescriptor, filetype: str, **kwargs):
        """
        Create a generic VTK file.

        :param filename: name of file or file handle
        :param filetype: VTK format of file
        :param **kwargs: parameters forwarded to Writer
        """
        self.filetype = filetype
        self.filename = filename
        self.writer = writer.Writer(filetype, **kwargs)

        # Center piece
        self.piece = self.writer.registerPiece()

        # Registering data elements
        self.point_data = self.piece.register('PointData')
        self.cell_data = self.piece.register('CellData')
        self.field_data = self.writer.registerComponent(
            'FieldData', self.writer.data_node
        )

    def addPointData(self, data_array: DataArray, vtk_format: str = 'binary'):
        """
        Add a DataArray instance to the PointData section of the file.

        :param data_array: DataArray instance
        :param vtk_format: data format. Can be:
            - ``'ascii'``: data is written with numpy.savetxt in-place
            - ``'binary'``: data is written in base64 (with possible
            compression) in-place
            - ``'append'``: data is written in base64 (with possible
            compression) in the ``AppendData`` section of the file
        """
        self.point_data.registerDataArray(data_array, vtk_format)
        return self

    def addCellData(self, data_array: DataArray, vtk_format: str = 'binary'):
        """
        Add a DataArray instance to the PointData section of the file.

        Arguments are identical to `addPointData`.
        """
        self.cell_data.registerDataArray(data_array, vtk_format)
        return self

    def addFieldData(self, data_array: DataArray, vtk_format: str = 'binary'):
        """
        Add a DataArray instance to the FieldData section of the file.

        Arguments are identical to `addPointData`.

        The FieldData section contains arrays that are not related to
        the domain geometry. Their shape is therefore freeform.
        """
        self.field_data.registerDataArray(data_array, vtk_format)
        return self

    def write(self):
        """Write XML to file."""
        self.writer.registerAppend()
        self.writer.write(self.filename)


class ImageData(VTKFile):
    """VTK Image data (coordinates given by a range and constant spacing)."""

    def __init__(self,
                 filename: VTKFile._FileDescriptor,
                 ranges: ts.List[ts.Tuple[float, float]],
                 points: ts.List[int],
                 offsets: ts.List[int] = None, **kwargs):
        """
        Init an ImageData file (regular orthogonal grid).

        :param filename: name of file or file handle
        :param ranges: list of pairs for coordinate ranges
        :param points: list of number of points
        """
        VTKFile.__init__(self, filename, 'ImageData', **kwargs)

        # Computing spacings
        spacings = [(x[1] - x[0]) / (n - 1) for x, n in zip(ranges, points)]

        if offsets is None:
            offsets = [0] * len(points)

        # Filling in missing coordinates
        points += [1] * max(3 - len(points), 0)
        offsets += [0] * max(3 - len(offsets), 0)

        # Setting extents, spacing and origin
        self.extent = _fold_extent([x - 1 for x in points],
                                   offsets, len(points))
        spacings = functools.reduce(lambda x, y: x + f"{y} ", spacings, "")
        origins = functools.reduce(lambda x, y: x + f"{y[0]} ", ranges, "")

        self.writer.setDataNodeAttributes({
            'WholeExtent': self.extent,
            'Spacing': spacings,
            'Origin': origins
        })

        self.piece.setAttributes({
            "Extent": self.extent
        })


class RectilinearGrid(VTKFile):
    """VTK Rectilinear grid (coordinates are given by 3 seperate arrays)."""

    def __init__(self,
                 filename: VTKFile._FileDescriptor,
                 coordinates: ts.Union[ts.Iterable[np.ndarray], np.ndarray],
                 offsets: ts.List[int] = None, **kwargs):
        """
        Init an RectilinearGrid file (irregular orthogonal grid).

        :param filename: name of file or file handle
        :param coordinates: list of coordinates for each direction
        """
        VTKFile.__init__(self, filename, 'RectilinearGrid', **kwargs)

        # Checking that we actually have a list or tuple
        if isinstance(coordinates, np.ndarray):
            coordinates = [coordinates]

        self.coordinates = list(coordinates)
        # Filling in missing coordinates
        self.coordinates += (
            [np.array([0.])] * max(0, 3 - len(self.coordinates))
        )

        # Setting data extent
        extent = []

        for coord in self.coordinates:
            if coord.ndim != 1:
                raise ValueError(
                    'Coordinate array should have only one dimension'
                    f' (has {coord.ndim})')
            extent.append(coord.size-1)

        # Create extent string with offsets
        self.extent = _fold_extent(extent, offsets, len(coordinates))

        self.writer.setDataNodeAttributes({
            "WholeExtent": self.extent
        })

        self.piece.setAttributes({
            "Extent": self.extent
        })

        # Registering coordinates
        coordinate_component = self.piece.register('Coordinates')

        for coord, prefix in zip(self.coordinates, ('x', 'y', 'z')):
            array = DataArray(coord, [0], prefix + '_coordinates')
            coordinate_component.registerDataArray(array, vtk_format='append')


class StructuredGrid(VTKFile):
    """VTK Structured grid (coordinates given by a single array of points)."""

    def __init__(self,
                 filename: VTKFile._FileDescriptor,
                 points: np.ndarray,
                 shape: ts.List[int], **kwargs):
        """
        Init a StructuredGrid file (ordered quadrangle/hexahedron cells).

        :param filename: name of file or file handle
        :param points: 2D numpy array of point coordinates
        :param shape: number of points in each spatial direction
        """
        VTKFile.__init__(self, filename, 'StructuredGrid', **kwargs)

        if points.ndim != 2:
            raise ValueError('Points should be a 2D array')

        # Completing the missing coordinates
        points = _make_3darray(points)

        extent = [n - 1 for n in shape]
        extent += [0] * max(3 - len(extent), 0)

        extent = functools.reduce(lambda x, y: x + f"0 {y} ", extent, "")
        self.writer.setDataNodeAttributes({
            "WholeExtent": extent
        })

        self.piece.setAttributes({
            "Extent": extent
        })

        points_component = self.piece.register('Points')
        points_component.registerDataArray(
            DataArray(points, [0], 'points'), vtk_format='append',
        )


class UnstructuredGrid(VTKFile):
    """VTK Unstructured grid (data on nodes + connectivity)."""

    def __init__(self,
                 filename: VTKFile._FileDescriptor,
                 nodes: np.ndarray,
                 connectivity: ts.Mapping[int, np.ndarray], **kwargs):
        """
        Init an UnstructuredGrid file (mesh with connectivity).

        :param filename: name of file or file handle
        :param nodes: 2D numpy array of node coordinates
        :param connectivity: dict with arrays for each cell type
        """
        VTKFile.__init__(self, filename, 'UnstructuredGrid', **kwargs)

        if nodes.ndim != 2:
            raise ValueError('Nodes should be a 2D array')

        # Completing the missing coordinates
        nodes = _make_3darray(nodes)

        if not check_connectivity(connectivity):
            raise ValueError('Connectivity is invalid')

        cells_component = self.piece.register('Cells')

        ncells = sum(map(len, connectivity.values()))

        self.piece.setAttributes({
            'NumberOfPoints': str(nodes.shape[0]),
            'NumberOfCells': str(ncells),
        })

        self.piece.register('Points').registerDataArray(
            DataArray(nodes, [0], 'points'), vtk_format='append'
        )

        int32 = np.dtype('i4')

        flat_connectivity = np.empty(
            sum(len(cell) for x in connectivity.values() for cell in x),
            dtype=int32,
        )

        # Flattening the connectivities for each element type
        offset = 0
        for conn in connectivity.values():
            for cell in conn:
                flat_connectivity[offset:offset+len(cell)] = cell
                offset += len(cell)

        offsets = np.empty(ncells, dtype=int32)

        offset, index = 0, 0
        for conn in connectivity.values():
            for cell in conn:
                offset += len(cell)
                offsets[index] = offset
                index += 1

        types = np.empty(ncells, dtype=int32)

        offset = 0
        for k, conn in connectivity.items():
            if isinstance(k, CellType):
                k = k.value
            types[offset:offset+len(conn)] = k
            offset += len(conn)

        # Register mesh description
        connectivity_data = {
            "connectivity": flat_connectivity,
            "offsets": offsets,
            "types": types,
        }

        for label, array in connectivity_data.items():
            cells_component.registerDataArray(
                DataArray(array, [0], label),
                vtk_format='append',
            )


class ParaViewData(WriteManager):
    """
    Groups VTK files into a single description.

    See: https://www.paraview.org/Wiki/ParaView/Data_formats#PVD_File_Format
    """

    def __init__(self, filename: str, **kwargs):
        """
        Initialize a PVD file.

        PVD files contain references to other data files, and are convenient
        to represent time series, for example.
        """
        self.filename = filename
        self.writer = writer.Writer('Collection', **kwargs)

    def addFile(self, file, timestep=0, group="", part=0):
        """
        Add a file to the group.

        :param file: filename or VTKFile instance
        :param timestep: real-time value of file
        :param group: group to add the file to
        :param part: sub-part of the domain represented by file
        """
        if isinstance(file, VTKFile):
            file = file.filename

        self.writer.registerComponent('DataSet', self.writer.data_node, dict(
            timestep=str(timestep), group=group, part=str(part), file=str(file)
        ))

    def write(self):
        """Write file."""
        self.writer.write(self.filename)



"""
Compatibility package with the high-level API of PyEVTK.

This package is meant as a drop-in replacement for high-level functions of
PyEVTK. However, it does not reproduce exactly the behavior of PyEVTK, most
notably:

- uvw.Writer keyword arguments (such as compression=True) can be passed to
  the functions
- Data fields with any number of components are accepted. In PyEVTK, data
  fields have to be either scalar or 3D by specifying a tuple of 3 Numpy
  arrays. This is inconvenient and inefficient. UVW's native DataArray object
  does not impose such requirements.
- The function imageToVTK() does not impose all dimensions to be of equal size
- Requirements on data are not enforced with assertions
- Data is written in base64 format instead of raw binary

Original version of PyEVTK by Paulo A. Herrera can be found at:

- https://github.com/paulo-herrera/PyEVTK
- https://github.com/pyscience-projects/pyevtk (corresponding to PyPI version)

Below is the copyright notice for PyEVTK:

MIT License

Copyright (c) 2010-2021 Paulo A. Herrera

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

"""

__copyright__ = "Copyright © 2018-2023 Lucas Frérot"
__license__ = "SPDX-License-Identifier: MIT"

from .. import vtk_files as vtk
from ..data_array import DataArray

import typing as ts
import numpy as np
from collections.abc import Sequence


DataType = ts.Mapping[
    str, ts.Union[np.ndarray,
                  ts.Tuple[np.ndarray, np.ndarray, np.ndarray]]
]


def _assemble(data):
    if isinstance(data, Sequence):
        assembled_data = np.zeros(list(data[0].shape) + [3],
                                  dtype=data[0].dtype)
        for i in range(3):
            assembled_data[..., i] = data[i]
        return assembled_data
    return data


def _add_data(cellData, pointData, fieldData, fh, dims=range(3)):
    valid_data = zip([cellData, pointData, fieldData],
                     ['addCellData', 'addPointData', 'addFieldData'])
    valid_data = filter(lambda d: d[0] is not None, valid_data)
    for data, m in valid_data:
        for k, v in data.items():
            v = _assemble(v)
            getattr(fh, m)(DataArray(v, dims, k))


def imageToVTK(
        path: str,
        origin: ts.Tuple[float, float, float] = (0.0, 0.0, 0.0),
        spacing: ts.Tuple[float, float, float] = (1.0, 1.0, 1.0),
        cellData: DataType = None,
        pointData: DataType = None,
        fieldData: DataType = None,
        **kwargs
) -> str:
    """
    Write data to an ImageData file.

    :param path: path to file without extension
    :param origin: 3-tuple giving the domain origin
    :param spacing: 3-tuple giving the point spacing
    :param cellData: dict of data defined on cells
    :param pointData: dict of data defined on points
    :param fieldData: dict of data with arbitrary support
    :param **kwargs: passed on to uvw.Writer (e.g. compression=True)
    :returns: full path of created file

    Notes:
    - domain shape is inferred from either cellData or pointData
    - unlike PyEVTK arrays do not have to have the same dimension in each
      direction
    """
    if cellData is None and pointData is None:
        raise ValueError("Cannot infer image size if cell "
                         "and point data are absent")

    def _extract_shape(value, offset, name):
        value = next(iter(value.values()))
        if isinstance(value, np.ndarray) and value.ndim == 3:
            return [s + offset for s in value.shape]
        if isinstance(value, Sequence) and value[0].ndim == 3:
            return [s + offset for s in value[0].shape]
        raise RuntimeError(f"Invalid dimension of {name}")

    # Deducing image size
    if cellData:
        shape = _extract_shape(cellData, 1, 'cellData')
    elif pointData:
        shape = _extract_shape(pointData, 0, 'pointData')

    ranges = [(o, o + n * s) for o, n, s in zip(origin, shape, spacing)]
    filename = path + '.vti'

    with vtk.ImageData(filename, ranges, shape, **kwargs) as fh:
        _add_data(cellData, pointData, fieldData, fh)
    return filename


def rectilinearToVTK(
        path: str,
        x: np.ndarray,
        y: np.ndarray,
        z: np.ndarray,
        cellData: DataType = None,
        pointData: DataType = None,
        fieldData: DataType = None,
        **kwargs
) -> str:
    """Write data to a RectilinearGrid.

    :param path: path to file without extension
    :param x: 1d array of x coordinates
    :param y: 1d array of y coordinates
    :param z: 1d array of z coordinates
    :param cellData: dict of data defined on cells
    :param pointData: dict of data defined on points
    :param fieldData: dict of data with arbitrary support
    :param **kwargs: passed on to uvw.Writer (e.g. compression=True)
    :returns: full path of created file

    """
    filename = path + '.vtr'
    with vtk.RectilinearGrid(filename, (x, y, z), **kwargs) as fh:
        _add_data(cellData, pointData, fieldData, fh)
    return filename


def structuredToVTK(
        path: str,
        x: np.ndarray,
        y: np.ndarray,
        z: np.ndarray,
        cellData: DataType = None,
        pointData: DataType = None,
        fieldData: DataType = None,
        **kwargs
) -> str:
    """Write data to a RectilinearGrid.

    :param path: path to file without extension
    :param x: 3d array of x coordinates
    :param y: 3d array of y coordinates
    :param z: 3d array of z coordinates
    :param cellData: dict of data defined on cells
    :param pointData: dict of data defined on points
    :param fieldData: dict of data with arbitrary support
    :param **kwargs: passed on to uvw.Writer (e.g. compression=True)
    :returns: full path of created file

    """
    filename = path + '.vts'
    points = np.vstack([x.ravel(), y.ravel(), z.ravel()]).T
    shape = [x.shape[0], y.shape[0], z.shape[0]]

    with vtk.StructuredGrid(filename, points, shape, **kwargs) as fh:
        _add_data(cellData, pointData, fieldData, fh)
    return filename


def gridToVTK(
        path: str,
        x: np.ndarray,
        y: np.ndarray,
        z: np.ndarray,
        cellData: DataType = None,
        pointData: DataType = None,
        fieldData: DataType = None,
        **kwargs
) -> str:
    """
    Write data to either a RectilinearGrid or StructuredGrid file.

    :param path: path to file without extension
    :param x: {1,3}-d array of x coordinates
    :param y: {1,3}-d array of y coordinates
    :param z: {1,3}-d array of z coordinates
    :param cellData: dict of data defined on cells
    :param pointData: dict of data defined on points
    :param fieldData: dict of data with arbitrary support
    :param **kwargs: passed on to uvw.Writer (e.g. compression=True)
    :returns: full path of created file

    Notes:
    - file type is infered from dimension of coordinate arrays:
      - 1D -> RectilinearGrid
      - 3D -> StructuredGrid
      3D grid coordinates can be obtained with numpy.meshgrid
    """
    dims = [x.ndim, y.ndim, z.ndim]

    if dims == [1] * 3:
        return rectilinearToVTK(path, x, y, z,
                                cellData, pointData, fieldData, **kwargs)
    if dims == [3] * 3:
        return structuredToVTK(path, x, y, z,
                               cellData, pointData, fieldData, **kwargs)

    raise ValueError(f"Wrong dimensions for x, y, z arrays: {dims}")


def pointsToVTK(
        path: str,
        x: np.ndarray,
        y: np.ndarray,
        z: np.ndarray,
        data: DataType = None,
        fieldData: DataType = None,
        **kwargs
) -> str:
    """
    Write point set to UnstructuredGrid file.

    :param path: path to file without extension
    :param x: 1d array of x coordinates
    :param y: 1d array of y coordinates
    :param z: 1d array of z coordinates
    :param data: dict of data defined on points
    :param fieldData: dict of data with arbitrary support
    :param **kwargs: passed on to uvw.Writer (e.g. compression=True)
    :returns: full path of created file
    """
    filename = path + '.vts'

    coords = np.vstack([x.ravel(), y.ravel(), z.ravel()]).T
    connectivity = {1: np.arange(coords.shape[0], dtype=np.int)}

    with vtk.UnstructuredGrid(filename, coords, connectivity, **kwargs) as fh:
        _add_data(None, data, fieldData, fh, [0])
    return filename


def linesToVTK(
        path: str,
        x: np.ndarray,
        y: np.ndarray,
        z: np.ndarray,
        cellData: DataType = None,
        pointData: DataType = None,
        fieldData: DataType = None,
        **kwargs
) -> str:
    """
    Write segment set to an UnstructuredGrid file.

    :param path: path to file without extension
    :param x: {1,3}-d array of x coordinates
    :param y: {1,3}-d array of y coordinates
    :param z: {1,3}-d array of z coordinates
    :param cellData: dict of data defined on cells
    :param pointData: dict of data defined on points
    :param fieldData: dict of data with arbitrary support
    :param **kwargs: passed on to uvw.Writer (e.g. compression=True)
    :returns: full path of created file

    Notes:
    - number of points should be even
    - in the final file, no point will appear twice in connectivity
      (i.e. the lines are all disconected)

    The above requirements are from pyevtk and are not restrictions of the file
    format. Use uvw.UnstructuredGrid for more flexibility.
    """
    points = np.vstack([x.ravel(), y.ravel(), z.ravel()]).T
    if points.shape[0] % 2 != 2:
        raise ValueError('Number of points should be even')

    connectivity = {
        3: np.arange(points.shape[0], dtype=np.int).reshape(points.shape[0], 2)
    }

    filename = path + '.vtu'
    with vtk.UnstructuredGrid(filename, points, connectivity, **kwargs) as fh:
        _add_data(cellData, pointData, fieldData, fh)
    return filename


def polyLinesToVTK(
        path: str,
        x: np.ndarray,
        y: np.ndarray,
        z: np.ndarray,
        pointsPerLine: np.ndarray,
        cellData: DataType = None,
        pointData: DataType = None,
        fieldData: DataType = None,
        **kwargs
) -> str:
    """
    Write segmented lines set to an UnstructuredGrid file.

    :param path: path to file without extension
    :param x: {1,3}-d array of x coordinates
    :param y: {1,3}-d array of y coordinates
    :param z: {1,3}-d array of z coordinates
    :param pointsPerLine: number of points per segmented line
    :param cellData: dict of data defined on cells
    :param pointData: dict of data defined on points
    :param fieldData: dict of data with arbitrary support
    :param **kwargs: passed on to uvw.Writer (e.g. compression=True)
    :returns: full path of created file
    """
    points = np.vstack([x.ravel(), y.ravel(), z.ravel()]).T
    start_ids = np.add.accumulate(pointsPerLine, dtype=np.int)
    connectivity = {4: np.array([
        np.arange(N, dtype=np.int) + s
        for N, s in zip(pointsPerLine, start_ids)
    ])}

    filename = path + '.vtu'
    with vtk.UnstructuredGrid(filename, points, connectivity, **kwargs) as fh:
        _add_data(cellData, pointData, fieldData, fh)
    return filename


def unstructuredGridToVTK(
        path: str,
        x: np.ndarray,
        y: np.ndarray,
        z: np.ndarray,
        connectivity: np.ndarray,
        offsets: np.ndarray,
        cell_types: np.ndarray,
        cellData: DataType = None,
        pointData: DataType = None,
        fieldData: DataType = None,
        **kwargs
) -> str:
    """Write UnstructuredGrid file.

    :param path: path to file without extension
    :param x: 1d array of x coordinates
    :param y: 1d array of y coordinates
    :param z: 1d array of z coordinates
    :param connectivity: 1d array of nodes for each element
    :param offsets: 1d array giving the position in connectivity of the first
                    element of each cell
    :param cell_types: 1d array giving the type of each cell
    :param cellData: dict of data defined on cells
    :param pointData: dict of data defined on points
    :param fieldData: dict of data with arbitrary support
    :param **kwargs: passed on to uvw.Writer (e.g. compression=True)
    :returns: full path of created file

    Note: this implementation does not use uvw.UnstructuredGrid because the
    function expects connectivity data in a format that uvw.UnstructuredGrid
    already converts to.

    """
    filename = path + '.vtu'
    points = np.vstack([x.ravel(), y.ravel(), z.ravel()]).T

    ncells = cell_types.size
    npoints = points.shape[0]

    with vtk.VTKFile(filename, 'UnstructuredGrid', **kwargs) as fh:
        fh.piece.settAttributes({
            'NumberOfPoints': str(npoints),
            'NumberOfCells': str(ncells),
        })

        fh.piece.register('Points').registerDataArray(
            DataArray(points, [0], 'points'),
        )

        cells_component = fh.register('Cells')
        connectivity_data = {
            "connectivity": connectivity,
            "offsets": offsets,
            "types": cell_types,
        }

        for label, array in connectivity_data.items():
            cells_component.registerDataArray(
                DataArray(array, [0], label),
            )

        _add_data(cellData, pointData, fieldData, fh)


def cylinderToVTK(
        path: str,
        x0: float,
        y0: float,
        z0: float,
        z1: float,
        radius: float,
        nlayers: int,
        npilars: int = 16,
        cellData: DataType = None,
        pointData: DataType = None,
        fieldData: DataType = None,
        **kwargs
):
    """Not yet implemented."""
    raise NotImplementedError()



"""UVW: Universal VTK Writer.

UVW is a small utility library to write XML VTK files from data contained in
Numpy arrays. It handles fully-fledged `ndarrays` defined over {1, 2, 3}-d
domains, with arbitrary number of components. There are no constraints on the
particular order of components, although copy of data can be avoided if the
array is Fortran contiguous, as VTK files are written in Fortran order. UVW
supports multi-process writing of VTK files, so that it can be used in an MPI
environment.

For documentation and examples, check out the various classes' docstrings and
the repository home on GitHub: https://github.com/prs513rosewood/uvw
"""

__author__ = "Lucas Frérot"
__maintainer__ = __author__
__copyright__ = "Copyright © 2018-2023 Lucas Frérot"
__license__ = "SPDX-License-Identifier: MIT"

# flake8: noqa

from .vtk_files import (
    RectilinearGrid,
    ImageData,
    StructuredGrid,
    UnstructuredGrid,
    ParaViewData
)

from .data_array import DataArray

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



"""Defining class (`DataArray`) used to represent Numpy arrays in XML model."""

__copyright__ = "Copyright © 2018-2023 Lucas Frérot"
__license__ = "SPDX-License-Identifier: MIT"

import functools
import typing as ts
import numpy as np

DTYPE_TO_VTK = {
    np.dtype('<f4'): 'Float32',
    np.dtype('<f8'): 'Float64',
    np.dtype('<i1'): 'Int8',
    np.dtype('<i2'): 'Int16',
    np.dtype('<i4'): 'Int32',
    np.dtype('<i8'): 'Int64',
    np.dtype('<u1'): 'UInt8',
    np.dtype('<u2'): 'UInt16',
    np.dtype('<u4'): 'UInt32',
    np.dtype('<u8'): 'UInt64',

    # Big-Endian variants (>i1 == <i1 and >u1 == <u1)
    np.dtype('>f4'): 'Float32',
    np.dtype('>f8'): 'Float64',
    np.dtype('>i2'): 'Int16',
    np.dtype('>i4'): 'Int32',
    np.dtype('>i8'): 'Int64',
    np.dtype('>u2'): 'UInt16',
    np.dtype('>u4'): 'UInt32',
    np.dtype('>u8'): 'UInt64',
}


class DataArray:
    """Class holding information on array data.

    This class allows ndarrays with arbitrary shape, including the number of
    components, to be used by the sub-classes of VTKFile. This is done by
    supplying the set of axis of the array that represent data in physical
    directions (e.g. x, y, z). This means that arrays defined on a {1,2,3}-d
    grid with N number of components are natively handled without having to
    reshape the array prior to writing the file.

    The VTK XML format assumes arrays are stored in Fortran order. By default,
    ndarrays are stored in C order, so a copy of the data may be necessary when
    the object is created. DataArray objects can also be told that components
    may be stored in C or Fortan order (e.g. if the data stored per point is a
    matrix).

    """

    def __init__(self,
                 data: ts.Sequence,
                 spatial_axes: ts.Iterable[int],
                 name: str = '',
                 components_order: str = 'C'):
        """
        Construct a data array.

        :param data: the numpy array containing the data (possibly a view)
        :param spatial_axes: a sequence of ints that indicate which axes of the
                             array correspond to space dimensions (in order)
        :param name: the name of the data
        :param components_order: the order of the non-spatial axes of the array
        """
        self.data = np.asanyarray(data)
        axes = list(range(self.data.ndim))
        spatial_axes = list(spatial_axes)

        if self.data.ndim < len(spatial_axes):
            raise ValueError(
                'Dimensions of data smaller than space dimensions')

        for ax in spatial_axes:
            axes.remove(ax)

        nb_components = functools.reduce(
            lambda x, y: x * self.data.shape[y], axes, 1)

        if components_order == 'C':
            axes.reverse()
        elif components_order == 'F':
            pass
        else:
            raise ValueError('Unrecognized components order')

        axes += spatial_axes

        # Hopefully this is a view
        self.flat_data = self.data.transpose(*axes).reshape(-1, order='F')

        try:
            data_type = DTYPE_TO_VTK[self.flat_data.dtype]
        except KeyError:
            raise TypeError(
                f'Array dtype {self.flat_data.dtype} is not supported by VTK')

        self.attributes = {
            "Name": name,
            "type": data_type,
            "NumberOfComponents": str(nb_components),
            "NumberOfTuples": str(self.flat_data.size),
        }

        self.format_str = '%d' if 'Int' in data_type else '%.18e'

    def __str__(self) -> str:  # pragma: no cover
        """Produce string representation."""
        return self.attributes.__str__()



"""Module with class for cell types in vtkUnstructuredGrid."""

__copyright__ = "Copyright © 2018-2023 Lucas Frérot"
__license__ = "SPDX-License-Identifier: MIT"

from enum import Enum, unique

import numpy as np

from .data_array import DTYPE_TO_VTK


@unique
class CellType(Enum):
    """
    Enumerates the VTK cell types.

    See https://kitware.github.io/vtk-examples/site/VTKFileFormats/#legacy-file-examples
    """

    VERTEX = 1
    POLY_VERTEX = 2
    LINE = 3
    POLY_LINE = 4
    TRIANGLE = 5
    TRIANGLE_STRIP = 6
    POLYGON = 7
    PIXEL = 8
    QUAD = 9
    TETRA = 10
    VOXEL = 11
    HEXAHEDRON = 12
    WEDGE = 13
    PYRAMID = 14
    PENTAGONAL_PRISM = 15
    HEXAGONAL_PRISM = 16
    QUADRATIC_EDGE = 21
    QUADRATIC_TRIANGLE = 22
    QUADRATIC_QUAD = 23
    QUADRATIC_TETRA = 24
    QUADRATIC_HEXAHEDRON = 25
    QUADRATIC_WEDGE = 26
    QUADRATIC_PYRAMID = 27
    BIQUADRATIC_QUAD = 28
    TRIQUADRATIC_HEXAHEDRON = 29
    QUADRATIC_LINEAR_QUAD = 30
    QUADRATIC_LINEAR_WEDGE = 31
    BIQUADRATIC_QUADRATIC_WEDGE = 32
    BIQUADRATIC_QUADRATIC_HEXAHEDRON = 33
    BIQUADRATIC_TRIANGLE = 34
    CUBIC_LINE = 35
    QUADRATIC_POLYGON = 36


NODES_PER_CELL = {
    CellType.VERTEX: 1,
    CellType.POLY_VERTEX: -1,
    CellType.LINE: 2,
    CellType.POLY_LINE: -1,
    CellType.TRIANGLE: 3,
    CellType.TRIANGLE_STRIP: -1,
    CellType.POLYGON: -1,
    CellType.PIXEL: 4,
    CellType.QUAD: 4,
    CellType.TETRA: 4,
    CellType.VOXEL: 8,
    CellType.HEXAHEDRON: 8,
    CellType.WEDGE: 6,
    CellType.PYRAMID: 5,
    CellType.PENTAGONAL_PRISM: 10,
    CellType.HEXAGONAL_PRISM: 12,
    CellType.QUADRATIC_EDGE: 3,
    CellType.QUADRATIC_TRIANGLE: 6,
    CellType.QUADRATIC_QUAD: 8,
    CellType.QUADRATIC_TETRA: 10,
    CellType.QUADRATIC_HEXAHEDRON: 20,
    CellType.QUADRATIC_WEDGE: 15,
    CellType.QUADRATIC_PYRAMID: 13,
    CellType.BIQUADRATIC_QUAD: 9,
    CellType.TRIQUADRATIC_HEXAHEDRON: 27,
    CellType.QUADRATIC_LINEAR_QUAD: 6,
    CellType.QUADRATIC_LINEAR_WEDGE: 12,
    CellType.BIQUADRATIC_QUADRATIC_WEDGE: 18,
    CellType.BIQUADRATIC_QUADRATIC_HEXAHEDRON: 24,
    CellType.BIQUADRATIC_TRIANGLE: 7,
    CellType.CUBIC_LINE: 4,
    CellType.QUADRATIC_POLYGON: -1,
}


def check_connectivity(connectivity):
    """Sanity check for number of nodes per cell."""
    for cell_type, conn in connectivity.items():
        if not isinstance(cell_type, CellType):
            cell_type = CellType(cell_type)
        if not isinstance(conn, np.ndarray):
            raise TypeError("Connectivity needs to be of type numpy.ndarray")

        int_types = {
            dtype for dtype, label in DTYPE_TO_VTK.items() if 'Int' in label
        }

        if conn.dtype not in int_types | {np.dtype(object)}:
            raise TypeError("Connectivity dtype needs to be an integer type or"
                            "an object type for variable size cells")
        nnodes = NODES_PER_CELL[cell_type]

        if nnodes != -1 and nnodes != conn.shape[1]:
            return False
    return True



"""Module with MPI-empowered classes for parallel VTK file types."""

__copyright__ = "Copyright © 2018-2023 Lucas Frérot"
__license__ = "SPDX-License-Identifier: MIT"

import functools

from os import PathLike
from os.path import splitext, basename
from mpi4py import MPI

from . import writer
from . import vtk_files

from .data_array import DataArray


MASTER_RANK = 0


def _check_file_descriptor(fd):
    """Check path argument."""
    if not isinstance(fd, (str, PathLike)):
        raise TypeError(f'Expected path, got {fd}')


def _str_convert(ex):
    """Convert list of string extents to list of tuples w/ mins and maxs."""
    return list(zip(*((int(x) for x in s.split()) for s in ex)))


def _min_max_reduce(extents):
    """Reduce local extents to global extent by looking for mins and maxs."""
    mins = map(min, extents[0::2])
    maxs = map(max, extents[1::2])

    return functools.reduce(
        lambda x, y: x + f"{y[0]} {y[0]+y[1]} ",
        zip(mins, maxs),
        ""
    )


class PVTKFile(vtk_files.VTKFile):
    """Generic parallel VTK file."""

    def __init__(self, filename, pfiletype, comm=None, **kwargs):
        """
        Create a generic VTK file.

        :param filename: name of file (cannot be file handle)
        :param pfiletype: VTK format of file
        :param comm: MPI communicator (default: MPI.COMM_WORLD)
        :param **kwargs: parameters forwarded to Writer
        """
        _check_file_descriptor(filename)
        filename = str(filename)  # Ensure we have a string
        self.pfilename = filename
        self.pfiletype = pfiletype
        self.comm = MPI.COMM_WORLD if comm is None else comm
        self.rank = self.comm.Get_rank()

        # Construct name of piece file
        extension = splitext(filename)[1]
        piece_extension = extension.replace('p', '')
        self.piece_name_template = (
            filename.replace(extension, '.rank{rank}') + piece_extension
        )

        # We deliberatly do not call VTKFile.__init__

        self.pwriter = writer.Writer(self.pfiletype)
        self.ppoint_data, self.pcell_data, self.pfield_data = [
            self.pwriter.registerComponent(label, self.pwriter.data_node)
            for label in "PPointData PCellData PFieldData".split()
        ]

    def addPointData(self, array, *args, **kwargs):
        """Add point data."""
        if self.rank == MASTER_RANK:
            self.ppoint_data.registerPDataArray(array, *args, **kwargs)
        return vtk_files.VTKFile.addPointData(self, array, *args, **kwargs)

    def addCellData(self, array, *args, **kwargs):
        """Add cell data."""
        if self.rank == MASTER_RANK:
            self.pcell_data.registerPDataArray(array, *args, **kwargs)
        return vtk_files.VTKFile.addCellData(self, array, *args, **kwargs)

    def addFieldData(self, array, *args, **kwargs):
        """Add field data."""
        if self.rank == MASTER_RANK:
            self.pfield_data.registerPDataArray(array, *args, **kwargs)
        return vtk_files.VTKFile.addFieldData(self, array, *args, **kwargs)

    def write(self):
        """Write file."""
        vtk_files.VTKFile.write(self)
        if self.rank == MASTER_RANK:
            self.pwriter.write(self.pfilename)


class PImageData(PVTKFile, vtk_files.ImageData):
    """Image data parallel writer."""

    parent = vtk_files.ImageData

    def __init__(self, filename, ranges, points, offsets, comm=None, **kwargs):
        """
        Init an ImageData file (regular orthogonal grid).

        :param filename: name of file or file handle
        :param ranges: list of pairs for coordinate ranges
        :param points: list of number of points
        :param comm: MPI communicator (default: MPI.COMM_WORLD)
        """
        PVTKFile.__init__(self, filename, 'PImageData', comm=comm, **kwargs)
        self.parent.__init__(
            self,
            self.piece_name_template.format(rank=self.rank),
            ranges,
            points,
            offsets,
            **kwargs
        )

        # Gather local data extents
        extents = self.comm.gather(self.extent, root=MASTER_RANK)
        spacings = [(x[1] - x[0]) / (n - 1) for x, n in zip(ranges, points)]
        ranges = self.comm.gather(ranges, root=MASTER_RANK)

        # Ensure on master
        if not extents:
            return

        for rank, extent in enumerate(extents):
            self.pwriter.registerPiece({
                'Source': basename(self.piece_name_template.format(rank=rank)),
                'Extent': extent
            })

        extents = _min_max_reduce(_str_convert(extents))

        # Reducing domain ranges
        mins, maxs = [], []
        for r in ranges:
            mins.append(list(map(min, r)))
            maxs.append(list(map(max, r)))
        mins = list(zip(*mins))
        maxs = list(zip(*maxs))
        ranges = [(min(rmin), max(rmax)) for rmin, rmax in zip(mins, maxs)]

        spacings = functools.reduce(lambda x, y: x + f"{y} ", spacings, "")
        origins = functools.reduce(lambda x, y: x + f"{y[0]} ", ranges, "")

        self.pwriter.setDataNodeAttributes({
            'WholeExtent': extents,
            'Spacing': spacings,
            'Origin': origins
        })


class PRectilinearGrid(PVTKFile, vtk_files.RectilinearGrid):
    """Rectilinear grid parallel writer."""

    parent = vtk_files.RectilinearGrid

    def __init__(self, filename, coordinates, offsets, comm=None, **kwargs):
        """
        Init the parallel counterpart of a RectilinearGrid file.

        :param filename: name of file or file handle
        :param coordinates: list of coordinate arrays for this rank
        :param offset: offset in global dataset for this rank
        :param comm: MPI communicator (default: MPI.COMM_WORLD)
        """
        PVTKFile.__init__(self, filename, 'PRectilinearGrid',
                          comm=comm, **kwargs)

        # Name of piece file associated to rank
        self.parent.__init__(
            self,
            self.piece_name_template.format(rank=self.rank),
            coordinates,
            offsets,
            **kwargs
        )

        # Gather local data extents
        extents = self.comm.gather(self.extent, root=MASTER_RANK)

        # Ensure on master
        if not extents:
            return

        # Register coordinates
        pcoordinates = self.pwriter.registerComponent('PCoordinates',
                                                      self.pwriter.data_node)

        for rank, extent in enumerate(extents):
            self.pwriter.registerPiece({
                'Source': basename(self.piece_name_template.format(rank=rank)),
                'Extent': extent
            })

        # Convert string extents to list of integers
        extent = _min_max_reduce(_str_convert(extents))

        self.pwriter.setDataNodeAttributes({
            'WholeExtent': extent
        })

        for coord, prefix in zip(self.coordinates, ('x', 'y', 'z')):
            array = DataArray(coord, [0], prefix + '_coordinates')
            pcoordinates.registerPDataArray(array)



"""
Module with classes for interacting with the XML model underlying VTK files.

See https://kitware.github.io/vtk-examples/site/VTKFileFormats/ for format
description.
"""

__copyright__ = "Copyright © 2018-2023 Lucas Frérot"
__license__ = "SPDX-License-Identifier: MIT"

from .data_array import DataArray

import xml.dom.minidom as dom
import io
import zlib
import typing as ts

import numpy as np

from base64 import b64encode
from os import PathLike
from collections.abc import Mapping


def setAttributes(node: dom.Node, attributes: ts.Mapping[str, ts.Any]):
    """Set attributes of a node."""
    for item in attributes.items():
        node.setAttribute(*item)


def encodeArray(array: np.ndarray, level: int) -> str:
    """Encode array data and header in base64."""
    def compress(array):
        """Compress array with zlib. Returns header and compressed data."""
        raw_data = memoryview(array.tobytes())
        data_size = raw_data.nbytes

        max_block_size = 2**15

        # Enough blocks to span whole data
        nblocks = data_size // max_block_size + 1
        last_block_size = data_size % max_block_size

        # Compress regular blocks
        compressed_data = [
            zlib.compress(raw_data[i*max_block_size:(i+1)*max_block_size],
                          level)
            for i in range(nblocks-1)
        ]

        # Compress last (smaller) block
        compressed_data.append(
            zlib.compress(raw_data[-last_block_size:], level)
        )

        # Header data (cf https://vtk.org/Wiki/VTK_XML_Formats#Compressed_Data)
        header_dtype = np.dtype(array.dtype.byteorder + 'u4')
        usize = max_block_size
        psize = last_block_size
        csize = [len(x) for x in compressed_data]
        header = np.array([nblocks, usize, psize] + csize, dtype=header_dtype)
        return header.tobytes(), b"".join(compressed_data)

    def raw(array):
        """Return header and array data in bytes."""
        header_dtype = np.dtype(array.dtype.byteorder + 'u4')
        header = np.array([array.nbytes], dtype=header_dtype)
        return header.tobytes(), memoryview(array)

    data = raw(array) if level == 0 else compress(array)
    return "".join(b64encode(x).decode() for x in data)


class Component:
    """Generic component class capable of registering sub-components."""

    def __init__(self, name: str, parent_node: dom.Node, writer):
        """Construct from name, parent node and writer object."""
        self.writer = writer
        self.document = writer.document
        self.node = self.document.createElement(name)
        parent_node.appendChild(self.node)

    def setAttributes(self, attributes: ts.Mapping[str, ts.Any]):
        """Set the node attributes from dictionary."""
        setAttributes(self.node, attributes)

    def register(
            self,
            name: str,
            attributes: ts.Optional[ts.Mapping[str, ts.Any]] = None
    ) -> 'Component':
        """Register a sub-component."""
        if attributes is None:
            attributes = {}

        if not isinstance(attributes, Mapping):
            raise ValueError(
                f'Cannot register attributes of type {type(attributes)}')

        sub_component = Component(name, self.node, self.writer)
        setAttributes(sub_component.node, attributes)
        return sub_component

    def _addArrayNodeData(self,
                          data_array: DataArray,
                          component: 'Component',
                          vtk_format: str):
        if vtk_format == 'ascii':
            sstream = io.StringIO()
            np.savetxt(
                sstream,
                data_array.flat_data,
                newline=' ',
                fmt=data_array.format_str,
            )
            data_as_str = sstream.getvalue()
            # reduce(lambda x, y: x + str(y) + ' ', data_array.flat_data, "")
        elif vtk_format == 'binary':
            data_as_str = encodeArray(data_array.flat_data,
                                      self.writer.compression)
        elif vtk_format == 'append':
            self.writer.append_data_arrays[component] = data_array
            return
        else:
            raise ValueError(f'Unsupported VTK Format "{vtk_format}"')

        component.node.appendChild(self.document.createTextNode(data_as_str))

    def _registerArrayComponent(self,
                                array: DataArray,
                                name: str,
                                vtk_format: str):
        attributes = array.attributes
        attributes['format'] = vtk_format
        return self.register(name, attributes)

    def registerDataArray(self,
                          data_array: DataArray,
                          vtk_format: str = 'binary'):
        """Register a DataArray object."""
        component = self._registerArrayComponent(data_array,
                                                 'DataArray',
                                                 vtk_format)
        self._addArrayNodeData(data_array, component, vtk_format)

    def registerPDataArray(self,
                           data_array: DataArray,
                           vtk_format: str = 'binary'):
        """Register a DataArray object in p-file."""
        self._registerArrayComponent(data_array, 'PDataArray', vtk_format)


class Writer:
    """Generic XML handler for VTK files."""

    FileDescriptor = ts.Union[str, PathLike, io.TextIOBase, io.BufferedIOBase]

    def __init__(self, vtk_format: str,
                 compression: ts.Optional[ts.Union[bool, int]] = None,
                 vtk_version: str = '0.1',
                 byte_order: str = 'LittleEndian'):
        """
        Create an XML writer.

        :param vtk_format: format of VTK file
        :param compression: compression level (see zlib), True, False or None
        :param vtk_version: version number of VTK file
        :param byte_order: byte order of binary data
        """
        valid_orders = {"LittleEndian", "BigEndian"}
        if byte_order not in valid_orders:
            raise ValueError(
                f"Byte order '{byte_order}' invalid, "
                f"should be in {valid_orders}")

        self.document = dom.getDOMImplementation()  \
                           .createDocument(None, 'VTKFile', None)
        self.root = self.document.documentElement
        self.root.setAttribute('type', vtk_format)
        self.root.setAttribute('version', vtk_version)
        self.root.setAttribute('byte_order', byte_order)
        self.data_node = self.document.createElement(vtk_format)
        self.root.appendChild(self.data_node)
        self.size_indicator_bytes = np.dtype('u4').itemsize
        self.append_data_arrays = {}

        if compression is None or compression is False or compression == 0:
            self.compression = 0
        elif compression is True:
            self.compression = -1
        elif compression in list(range(-1, 10)):
            self.compression = compression
        else:
            raise ValueError(f'compression level {compression} is not '
                             'recognized by zlib')

        if self.compression != 0:
            self.root.setAttribute('compressor', 'vtkZLibDataCompressor')

    def setDataNodeAttributes(self, attributes: ts.Mapping[str, ts.Any]):
        """Set attributes for the entire dataset."""
        setAttributes(self.data_node, attributes)

    def registerPiece(self, attributes: ts.Mapping[str, ts.Any] = {}):
        """Register a piece element."""
        return self.registerComponent('Piece', self.data_node,
                                      attributes)

    def registerComponent(
            self,
            name: str,
            parent: dom.Node,
            attributes: ts.Mapping[str, ts.Any] = {}
    ) -> Component:
        """Register a Component to a parent Component with set attributes."""
        comp = Component(name, parent, self)
        setAttributes(comp.node, attributes)
        return comp

    def registerAppend(self):
        """Register AppendedData node (should only be called once)."""
        append_node = Component('AppendedData', self.root, self)
        append_node.setAttributes({'format': 'base64'})
        self.root.appendChild(append_node.node)

        data_str = "_"
        offset = 0
        for component, data in self.append_data_arrays.items():
            component.setAttributes({'offset': str(offset)})
            data_b64 = encodeArray(data.flat_data, self.compression)
            data_str += data_b64
            offset += len(data_b64)

        text = self.document.createTextNode(data_str)
        append_node.node.appendChild(text)

    def write(self, fd: FileDescriptor):
        """Write to file descriptor."""
        if isinstance(fd, (str, PathLike)):
            with open(fd, 'wb') as fh:
                self.write(fh)
        elif isinstance(fd, io.TextIOBase):
            self.document.writexml(fd, indent="\n  ", addindent="  ")
        elif isinstance(fd, io.BufferedIOBase):
            fd.write(self.document.toxml(encoding='UTF-8'))
        else:
            raise TypeError(f"Expected a path or file descriptor, got {fd}")

    def __str__(self) -> str:
        """Print XML to string."""
        return self.document.toprettyxml()



"Install script"
import setuptools
import versioneer

with open("README.md", "r") as fh:
    long_description = fh.read().replace('lang=', '')

setuptools.setup(
    name="uvw",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author="Lucas Frérot",
    author_email="lucas.frerot@protonmail.com",
    description="Universal VTK Writer for Numpy Arrays",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/prs513rosewood/uvw",
    packages=setuptools.find_packages(),
    install_requires=['numpy'],
    extras_require={
      "mpi": ['mpi4py'],
      "tests": ['pytest', 'pytest-mpi', 'pytest-cov', 'mpi4py', 'vtk'],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Visualization",
        "Intended Audience :: Science/Research",
    ],
)
